from pathlib import Path
from typing import Any, ClassVar
from datetime import date, datetime
from collections import OrderedDict
from io import BytesIO
from zipfile import ZipFile
import locale
import logging
from json import dumps

import numpy as np
import pandas as pd
import geopandas as gpd

import shapely
import pyproj

from sqlmodel import SQLModel, Field

from geoalchemy2.shape import from_shape

from sqlalchemy import BigInteger, Column, String, func, and_
from sqlalchemy.sql import sqltypes
from psycopg2.extensions import adapt

from geoalchemy2.types import Geometry
from geoalchemy2.elements import WKBElement

from shapely import wkb
from shapely.geometry import mapping
from shapely.ops import transform

from shapefile import (Writer as ShapeFileWriter,
                       POINT, POINTZ,
                       POLYLINE, POLYLINEZ,
                       POLYGON, POLYGONZ,
                       )

from ..config import conf
from .models_base import Model
from .survey import Equipment, Surveyor, Accuracy
from .misc import Qml
from .category import Category
from .project import Project
from ..utils import upsert_df

LOCALE_DATE_FORMAT = locale.nl_langinfo(locale.D_FMT)


## Coordinates of WGS84 should be in the (y, x) order, but some components (shapely?)
## always work with (x, y). Force (y, x) order with always_xy=True
## https://github.com/pyproj4/pyproj/issues/504
## https://proj.org/faq.html#why-is-the-axis-ordering-in-proj-not-consistent
reproject_func = pyproj.Transformer.from_crs(
    conf.geo.srid, conf.geo.srid_for_proj, always_xy=True).transform


class NoPoint(Exception):
    pass

logger = logging.getLogger('model_base_base')


## Mapping of sql types to shapefile types, for pyshp
## Character, Numbers, Longs, Dates, or Memo
exportable_cols = {
    sqltypes.Numeric: 'N',
    sqltypes.Integer: 'N',
    sqltypes.DateTime: 'D',
    sqltypes.Date: 'D',
    sqltypes.Time: 'D',
    sqltypes.String: 'C',
    sqltypes.Unicode: 'C',
}


class BaseSurveyModel(SQLModel):
    """
    Base mixin class for all layers defined from a category:
    - raw survey (RAW_V_*')
    - projected ('V_*')
    """
    id: int = Field(sa_column=Column(BigInteger()), primary_key=True)
    equip_id: int = Field(foreign_key='equipment.id')
    srvyr_id: int = Field('surveyor.id')
    accur_id: int = Field('accuracy.id')
    project_id: int = Field('project.id')

    orig_id: str
    date: date

    @classmethod
    def dyn_join_with(cls):
        return {
            'equipment': Equipment,
            'surveyor': Surveyor,
            'accuracy': Accuracy,
            'project': Project,
        }

    #async def get_info(self):
    #    info = await super(BaseSurveyModel, self).get_info()
    #    return info

    #async def get_geo_info(self):
    #    info = await super(BaseSurveyModel, self).get_geo_info()
    #    return info

    async def get_survey_info(self):
        info = await super(BaseSurveyModel, self).get_survey_info()
        if self.category:
            info['ISO layer name'] = self.iso_layer_name
            info['survey category'] = '{} ({})'.format(self.category.description, self.category.name)
        if self.project_id:
            info['project'] = self.project.name
        if self.srvyr_id:
            info['surveyor'] = self.surveyor.name
        if self.equip_id:
            info['survey equipment'] = self.equipment.name
        if self.accur_id:
            info['survey accuracy'] = self.accuracy.name
        if self.date:
            info['survey date'] = self.date.strftime(LOCALE_DATE_FORMAT)
        if self.orig_id:
            info['original id'] = self.orig_id
        return info

    @property
    def iso_layer_name(self):
        """
        The ISO layer name, built on the category and status
        """
        return '{category.domain}-{category.group:4s}-{category.minor_group_1:4s}-{category.minor_group_2:4s}-{status:1s}'\
            .format(category=self.category, status=self.status)


class SurveyModel(BaseSurveyModel):
    """
    Base mixin class for defining final (reprojected) survey data, with a status
    """
    status: str = Field(sa_column=Column(String(1)))

    get_gdf_with_related: bool = False

    filtered_columns_on_map: list[str] = [
        'equip_id',
        'srvyr_id',
        'accur_id',
        'project_id',
        'orig_id',
        'date',
        'gisaf_survey_equipment_id',
        'gisaf_survey_equipment_name',
        'gisaf_survey_surveyor_id',
        'gisaf_survey_surveyor_name',
        'gisaf_survey_accuracy_id',
        'gisaf_survey_accuracy_name',
        'gisaf_survey_accuracy_accuracy',
        'gisaf_admin_project_id',
        'gisaf_admin_project_name',
        'gisaf_admin_project_contact_person',
        'gisaf_admin_project_site',
        'gisaf_admin_project_date_approved',
        'gisaf_admin_project_start_date_planned',
        'gisaf_admin_project_start_date_effective',
        'gisaf_admin_project_end_date_planned',
        'gisaf_admin_project_end_date_effective',
        'gisaf_admin_project_skip_columns',
    ]

    async def get_survey_info(self):
        info = await super(SurveyModel, self).get_survey_info()
        if self.srvyr_id:
            info['surveyor'] = self.surveyor.name
        if self.equip_id:
            info['survey equipment'] = self.equipment.name
        if self.accur_id:
            info['survey accuracy'] = self.accuracy.name
        if self.date:
            info['survey date'] = self.date.strftime(LOCALE_DATE_FORMAT)
        info['status'] = self.status
        info['original id'] = self.orig_id
        return info

    async def get_feature_properties(self):
        return {
            'status': self.status if self.status else 'E'
        }

    @property
    def caption(self):
        return '{self.category.description} [{self.category.group}-{self.category.minor_group_1}] #{self.id:d}'.format(self=self)

    @classmethod
    async def get_popup(cls, df):
        return cls.category.description + \
            ' [' + cls.category.group + '-' + cls.category.minor_group_1 + \
            '] #' + df.index.astype('U')

    @classmethod
    async def get_geojson(cls, simplify_tolerance=0):

        ## Fastest, but the id is in properties and needs the front end (eg Gisaf for mapbox)
        ## to move it at the feature level
        ## Requires PostGis 3.0+
        sql_query_for_geojson_fast = """
        SELECT json_build_object(
            'type', 'FeatureCollection',
            'features', json_agg(ST_AsGeoJSON(t.*)::json)
        )::varchar
        FROM (
            SELECT f.geom,
                f.id::varchar,
                {description} || ' [' || '{model.category.group}' || '-' || '{model.category.minor_group_1}' || '] #' || f.id as popup,
                f.status
            FROM "{schema}"."{table}" as f
            WHERE f.geom is not null
        ) AS t;
        """

        ## Slower version (3x), but compliant with RFC 7946
        ## where the id is at the feature level
        sql_query_for_geojson_slow = """
        SELECT jsonb_build_object(
            'type', 'FeatureCollection',
            'features', jsonb_agg(features.feature)
        )::varchar
        FROM (
            SELECT jsonb_build_object(
                'type', 'Feature',
                'id', id,
                'geometry', ST_AsGeoJSON(geom)::jsonb,
                'properties', jsonb_build_object(
                    'popup',
                    {description} || ' [' || '{model.category.group}' || '-' || '{model.category.minor_group_1}' || '] #' || inputs.id::varchar,
                    'status', status
                )
            ) AS feature
            FROM "{schema}"."{table}" AS inputs
        ) AS features
        """

        sql_query_for_geojson = sql_query_for_geojson_fast

        async with db.acquire(reuse=False) as conn:
            query = sql_query_for_geojson.format(
                model=cls,
                schema=cls.__table_args__['schema'],
                table=cls.__tablename__,
                description=adapt(cls.category.description),
            )
            result = await conn.scalar(query)
            return result


    def to_row(self):
        """
        Get a list of attributes, typically used for exporting in CSV
        :return: list of attributes
        """
        return [
            self.id,
            self.shapely_geom.x,
            self.shapely_geom.y,
            self.shapely_geom.z,
            self.category_name,
            self.surveyor,
            self.equipment,
            self.date.isoformat(),
            self.accuracy.name,
            self.status,
            self.project.name,
            self.orig_id
        ]


class GeoModel(Model):
    """
    Base class for all geo models
    """
    #__abstract__ = True
    description: str = ''
    attribution: str | None = None

    can_get_features_as_df: bool = True
    """
    can_get_features_as_df indicates that the model is ready to get GeoJson using GeoDataframe
    If False, switch back to gino and dict based conversion using get_features_in_bulk_gino
    and record.get_feature_as_dict (DEPRECATED)
    """

    cache_enabled: bool = True
    """
    cache_enabled indicated that the model is OK with the caching mechanism of geojson stores.
    The cache is time-stamped with DB triggers on modification, so it's safe unless the model
    fetches other data than from its own table, eg. some status for styling.
    See gisaf.redis_tools and geoapi.gj_feature for the implementation details of the cache.
    """

    get_gdf_with_related: bool = False
    """
    get_gdf_with_related indicates that get_df (thus, get_geo_df and the geoJson API for
    the map online) gets related models (1-n relations, as defined with _join_with and dyn_join_with)
    by default.
    It can be overridden with the with_related parameter when calling get_df.
    """

    z_index: int = 450
    """
    z-index for the leaflet layer.
    Should be between 400 and 500.
    """

    icon: str | None = None
    """
    Icon for the model, used for normal web interface (ie. except the map)
    """

    symbol: str | None = None
    """
    Icon for the model, used in the map (mapbox)
    """

    style: str = ''
    """
    Style for the model, used in the map, etc
    """

    status: str = 'E'
    """
    Status (ISO layers definition) of the layer. E -> Existing.
    """

    _join_with: dict[str, Any] = {
    }
    """
    Fields to join when getching items using get_features.
    """

    hidden: bool = False
    """
    This model should be hidden from the menu
    """

    def __str__(self):
        return self.caption

    async def get_geo_info(self):
        """
        Geographical info
        """
        return {}

    async def get_survey_info(self):
        """
        Quality info: project, source, accuracy...
        """
        return OrderedDict()

    async def get_info(self):
        """
        Model specific info
        """
        return {}

    @classmethod
    def get_join_with(cls):
        if hasattr(cls, 'dyn_join_with'):
            return cls.dyn_join_with()
        else:
            return cls._join_with

    #@classmethod
    #def get_style(cls):
    #    """
    #    Style for the layer in Leaflet (css).
    #    :return:
    #    """
    #    return {
    #        # 'color': "#4B1BDE",
    #        'weight': '1'
    #    }

    @classmethod
    def get_options(cls):
        """
        Options for the layer in Leaflet.
        For example: {'weight': 0, 'fillOpacity': 0.1}
        :return:
        """
        return {
        }

    @property
    def caption(self):
        """
        Subclass me!
        :return: str
        """
        return 'Some geometry'

    @classmethod
    async def get_popup(cls, df):
        return cls.__name__ + ': ' + df.index.astype('U')

    @classmethod
    async def get_properties(cls, df):
        return {}

    async def get_tags(self):
        from gisaf.models.tags import Tags
        tags = await Tags.get_df(
            where=and_(
                Tags.store == self.__class__.get_store_name(),
                Tags.ref_id == self.id
            ),
            with_only_columns=['tags']
        )
        if len(tags) > 0:
            return tags.loc[0, 'tags']
        else:
            return {}

    @property
    def shapely_geom(self):
        if not hasattr(self, '_shapely_geom'):
            if isinstance(self.geom, WKBElement):
                bytes = self.geom.data
                if bytes:
                    self._shapely_geom = wkb.loads(bytes)
                else:
                    self._shapely_geom = None
            else:
                self._shapely_geom = None
        return self._shapely_geom

    def get_bgColor(self):
        """
        Get the background color of the element on the map.
        Subclass me!
        :return: str
        """
        return ''

    async def get_feature_as_dict(self, simplify_tolerance=None, reproject=False, css_class_prefix=''):
        """
        Get the parameters of this object (feature)
        :param css_class_prefix: for leaflet only
        :return:
        """
        logger.warn(f'{self} use get_feature_as_dict: please DEPRECATE in favor of get_geo_df')

        if hasattr(self, 'get_feature_properties'):
            properties = await self.get_feature_properties()
        else:
            properties = {
                'popup': self.caption
            }

        if reproject:
            shapely_geom = transform(reproject_func, self.shapely_geom)
        else:
            shapely_geom = self.shapely_geom

        if simplify_tolerance:
            shapely_geom = shapely_geom.simplify(simplify_tolerance / conf.geo['simplify_geom_factor'],
                                                 preserve_topology=False)
            if shapely_geom.is_empty:
                raise NoPoint

        return {
            "type": "Feature",
            'id': str(self.id),
            "properties": properties,
            'geometry': mapping(shapely_geom)
        }

    @classmethod
    def get_columns_for_shapefile_writer(cls, writer):
        """
        Return the columns and corresponding attribute name
        :param writer:
        :return:
        """

        ## Define the attributes
        columns = cls.__table__.columns
        cols = [col.name for col in columns if col.type.__class__ in exportable_cols]
        for col in columns:
            if col.name in cols:
                writer.field(col.name, exportable_cols[col.type.__class__])

        return cols

        return OrderedDict([(col, cols_to_attr[col]) for col in cols])

    @classmethod
    async def get_shapefile_writer(cls):
        """
        Get a shapefile writer with all records loaded.
        Subclasses should implement this.
        :return:
        """
        raise NotImplementedError

    @classmethod
    async def get_shapefile_files(cls):
        """
        Get an actual shapefile with all records.
        :return: zip file
        """
        writer = await cls.get_shapefile_writer()
        layer_name = cls.__name__

        ## Save the dbf, shp, shx data
        dbf = BytesIO()
        writer.saveDbf(dbf)
        shp = BytesIO()
        writer.saveShp(shp)
        shx = BytesIO()
        writer.saveShx(shx)

        qml = None
        qml_rec = await Qml.get(layer_name)
        if qml_rec and qml_rec.qml:
            qml = qml_rec.qml

        srid = cls.geom.expression.type.srid

        ## Write the .prj file: projection system reference
        ## Try if there's a file
        prj_path = (conf.prj_files_dir / str(srid)).with_suffix('.prj')
        proj_str = None
        if prj_path.is_file():
            with open(prj_path) as proj_file:
                proj_str = proj_file.read()
        else:
            ## Get from the spatial_ref_sys table in the database
            async with db.acquire(reuse=False) as connection:
                async with connection.transaction() as tx:
                    proj_rec = await connection.first(f'select srtext from spatial_ref_sys where srid={srid}')
            if proj_rec and proj_rec[0]:
                proj_str = proj_rec[0]

        return dbf, shp, shx, qml, proj_str

    @classmethod
    async def get_shapefile(cls):
        layer_name = cls.__name__
        dbf, shp, shx, qml, proj_str = await cls.get_shapefile_files()

        ## Zip and return data
        zip_file = BytesIO()
        with ZipFile(zip_file, 'w') as zip:
            zip.writestr('{}.dbf'.format(layer_name), dbf.getvalue())
            zip.writestr('{}.shp'.format(layer_name), shp.getvalue())
            zip.writestr('{}.shx'.format(layer_name), shx.getvalue())
            if qml:
                zip.writestr('{}.qml'.format(layer_name), qml)
            if proj_str:
                zip.writestr('{}.prj'.format(layer_name), proj_str)

        zip_file.seek(0)
        return zip_file

    @classmethod
    async def get_mapbox_style(cls):
        """
        Get the mapbox style (paint, layout, attribution...)
        """
        ## If the model is from survey, it should have a category, which has a style
        ## Get from database
        style = {}
        if hasattr(cls, 'category'):
            category = await Category.get_item_by_pk(pk=cls.category.name)
            if category:
                if category['mapbox_paint'] is not None:
                    style['paint'] = category['mapbox_paint']
                if category['mapbox_layout'] is not None:
                    style['layout'] = category['mapbox_layout']

        else:
            category = None

        qml = await Qml.get_item_by_pk(pk=cls.__name__)
        if qml:
            if qml['mapbox_paint']:
                style['paint'] = qml['mapbox_paint']
            if qml['mapbox_layout']:
                style['layout'] = qml['mapbox_layout']

        if cls.attribution is not None:
            style['attribution'] = cls.attribution
        return style


    @classmethod
    async def get_features_attrs(cls, simplify_tolerance):
        """
        Get the attributes and kwargs to pass to get_features for this model
        :param simplify_tolerance: float
        :return: tuple(dict, dict)
        """
        attrs = {"type": "FeatureCollection"}

        logger.warn(f'{cls} use get_features_attrs: please DEPRECATE in favor of get_geo_df')

        ## Get style, with priority order
        ## If the model is from survey ("auto_model"), it should have a category, which has a style
        category = getattr(cls, 'category', None)

        features_kwargs = {}

        if not issubclass(cls, GeoPointModel):
            if hasattr(cls, 'simplify') and cls.simplify:
                features_kwargs['simplify_tolerance'] = float(cls.simplify)
            else:
                features_kwargs['simplify_tolerance'] = simplify_tolerance
        features_kwargs['reproject'] = conf.geojson_srid != conf.srid

        ## If the model is from survey, it should have a category, which has a style
        ## Get from database
        ## TODO: update categories dynamically in the registry
        if hasattr(cls, 'category'):
            category = await Category.get(cls.category.name)
        else:
            category = None

        features_kwargs['reproject'] = conf.geojson_srid != conf.srid

        ## Get options, if the model has defined the get_options function
        if hasattr(cls, 'get_options'):
            attrs["options"] = cls.get_options()

        if hasattr(cls, 'get_popup_dynamic'):
            attrs["has_popups"] = True

        return attrs, features_kwargs

    @classmethod
    async def get_features_in_bulk_gino(cls, **kwargs):
        """
        Get the features, using joins defined by _join_with, or, in priority, get_join_with().
        See Gino's doc: relationship.html (specially loaders)
        :param kwargs:
        :return:
        """
        features = []
        async with db.acquire(reuse=True) as connection:
            async with connection.transaction() as tx:
                join = cls.get_join_with()
                query = cls.load(**join)
                #async for record in connection.iterate(query):
                for record in await connection.all(query):
                    if record.geom is not None:
                        try:
                            feature = await record.get_feature_as_dict(**kwargs)
                        except NoPoint:
                            pass
                        except Exception as err:
                            logger.exception(err)
                        else:
                            features.append(feature)

        return features

    @classmethod
    async def get_geos_df(cls, where=None, crs=None, reproject=False,
                         filter_columns=False, with_popup=False,
                         drop_wkb=True, **kwargs):
        if not gpd.options.use_pygeos:
            logger.warn(f'Using get_geos_df for {cls} but gpd.options.use_pygeos not set')
        if not crs:
            crs = conf.crs['geojson']
        df = await cls.get_df(where=where, **kwargs)
        df.set_index('id', inplace=True)
        df.rename(columns={'geom': 'wkb'}, inplace=True)
        df['geometry'] = shapely.from_wkb(df['wkb'])
        if drop_wkb:
            df = df.drop(columns=['wkb'])
        return gpd.GeoDataFrame(df, geometry='geometry', crs=crs)


    @classmethod
    async def get_geo_df(cls, where=None, crs=None, reproject=False,
                         filter_columns=False, with_popup=False, **kwargs):
        """
        Return a Pandas dataframe of all records
        :param where: where clause for the query (eg. Model.attr=='foo')
        :param crs: coordinate system (eg. 'epsg:4326') (priority over the reproject parameter)
        :param reproject: should reproject to conf.srid_for_proj
        :return:
        """
        df = await cls.get_df(where=where, **kwargs)
        df.dropna(subset=['geom'], inplace=True)
        df.set_index('id', inplace=True)
        df.sort_index(inplace=True)
        df_clean = df[df.geom != None]
        ## Drop coordinates
        df_clean.drop(columns=set(df_clean.columns).intersection(['ST_X_1', 'ST_Y_1', 'ST_Z_1']),
                      inplace=True)
        if not crs:
            crs = conf.crs['geojson']

        ## XXX: is it right? Waiting for application.py to remove pygeos support to know more.
        if getattr(gpd.options, 'use_pygeos', False):
            geometry = shapely.from_wkb(df_clean.geom)
        else:
            geometry = [wkb.loads(geom) for geom in df_clean.geom]

        gdf = gpd.GeoDataFrame(
            df_clean.drop('geom', axis=1),
            crs=crs,
            geometry=geometry
        )

        if hasattr(cls, 'simplify') and cls.simplify:
            #shapely_geom = shapely_geom.simplify(simplify_tolerance / conf.geo['simplify_geom_factor'],
                                                 #preserve_topology=False)
            gdf['geometry'] = gdf['geometry'].simplify(
                float(cls.simplify) / conf.geo['simplify_geom_factor'],
                preserve_topology=False)

        if reproject:
            gdf.to_crs(crs=conf.crs['for_proj'], inplace=True)

        ## Filter out columns
        if filter_columns:
            gdf.drop(columns=set(gdf.columns).intersection(cls.filtered_columns_on_map),
                     inplace=True)

        if with_popup:
            gdf['popup'] = await cls.get_popup(gdf)

        return gdf

    @classmethod
    def get_attachment_dir(cls):
        return f'{cls.__table__.schema}.{cls.__table__.name}'

    @classmethod
    def get_attachment_base_dir(cls):
        return Path(conf.attachments['base_dir'])/cls.get_attachment_dir()

class Geom(str):
    pass

class GeoPointModel(GeoModel):
    #__abstract__ = True
    shapefile_model: ClassVar[int] = POINT
    geom: Any = Field(sa_column=Column(Geometry('POINT', srid=conf.geo.srid)))
    icon: str | None = None
    mapbox_type: str = 'symbol'
    base_gis_type: str = 'Point'
    symbol: str = '\ue32b'

    @property
    def latitude(self):
        return self.shapely_geom.y

    @property
    def longitude(self):
        return self.shapely_geom.x

    @property
    def elevation(self):
        return self.shapely_geom.z

    @property
    def caption(self):
        """
        Return user friendly name (used in menu, etc)
        :return:
        """
        return '{self.__class__.__name__}: {self.id:d}'.format(self=self)

    def get_coords(self):
        return (self.shapely_geom.x, self.shapely_geom.y)

    @classmethod
    async def get_shapefile_writer(cls):
        writer = ShapeFileWriter(cls.shapefile_model)
        cols = cls.get_columns_for_shapefile_writer(writer)
        query = cls.query.where(cls.geom != None)

        async with db.acquire(reuse=False) as connection:
            async with connection.transaction() as tx:
                ## Add the objects
                async for obj in connection.iterate(query):
                    writer.point(*obj.get_coords(), shapeType=cls.shapefile_model)
                    attrs = []
                    for attr_name in cols:
                        attr = getattr(obj, attr_name)
                        if isinstance(attr, date):
                            ## Date in a format approriate for shapefile
                            attrs.append('{:%Y%m%d}'.format(attr))
                        else:
                            attrs.append(attr)
                    writer.record(*attrs)

        return writer

    async def get_geo_info(self):
        info = OrderedDict()
        if self.shapely_geom:
            info['longitude'] = '{:.6f}'.format(self.shapely_geom.x)
            info['latitude'] = '{:.6f}'.format(self.shapely_geom.y)
        return info


class GeoPointZModel(GeoPointModel):
    #__abstract__ = True
    geom: Any = Field(sa_column=Column(Geometry('POINTZ', dimension=3, srid=conf.geo.srid)))
    shapefile_model: int = POINTZ

    def get_coords(self):
        return (self.shapely_geom.x, self.shapely_geom.y, self.shapely_geom.z)

    async def get_geo_info(self):
        info = await super(GeoPointZModel, self).get_geo_info()
        info['elevation (m)'] = '{:.2f}'.format(self.shapely_geom.z)
        return info


class GeoPointMModel(GeoPointZModel):
    #__abstract__ = True
    shapefile_model: int = POINTZ
    geom: Any = Field(sa_column=Column(Geometry('POINTZ', dimension=3, srid=conf.geo.srid)))


class GeoLineModel(GeoModel):
    #__abstract__ = True
    shapefile_model: int = POLYLINE
    geom: Any = Field(sa_column=Column(Geometry('LINESTRING', srid=conf.geo.srid)))
    mapbox_type: str = 'line'
    base_gis_type: str = 'Line'

    @property
    def length(self):
        return transform(reproject_func, self.shapely_geom).length

    @property
    def caption(self):
        """
        Return user friendly name (used in menu, etc)
        :return:
        """
        return '{self.__class__.__name__}: {self.id:d}'.format(self=self)

    @classmethod
    async def get_shapefile_writer(cls):
        writer = ShapeFileWriter(cls.shapefile_model)
        cols = cls.get_columns_for_shapefile_writer(writer)
        query = cls.query.where(cls.geom != None)

        async with db.acquire(reuse=False) as connection:
            async with connection.transaction() as tx:
                ## Add the objects
                async for obj in connection.iterate(query):
                    try:
                        writer.line(parts=[[point for point in obj.get_points()]])
                    except NoPoint:
                        pass
                    attrs = []
                    for attr_name in cols:
                        attr = getattr(obj, attr_name)
                        if isinstance(attr, date):
                            ## Date in a format approriate for shapefile
                            attrs.append('{:%Y%m%d}'.format(attr))
                        else:
                            attrs.append(attr)
                    writer.record(*attrs)

        return writer

    def get_points(self):
        """
        Get a list of points
        :return:
        """
        if isinstance(self.geom, None.__class__):
            raise NoPoint
        points = wkb.loads(self.geom.data)
        return zip(points.coords.xy[0], points.coords.xy[1])

    async def get_geo_info(self):
        info = OrderedDict()
        bounds = self.shapely_geom.bounds
        info['longitude'] = '{:.6f} - {:.6f}'.format(bounds[0], bounds[2])
        info['latitude'] = '{:.6f} - {:.6f}'.format(bounds[1], bounds[3])
        info['length (m)'] = '{self.length:.2f}'.format(self=self)
        return info


class GeoLineModelZ(GeoLineModel):
    #__abstract__ = True
    shapefile_model: int = POLYLINEZ
    geom: Any = Field(sa_column=Column(Geometry('LINESTRINGZ', dimension=3, srid=conf.geo.srid)))

    async def get_geo_info(self):
        info = await super(GeoLineModelZ, self).get_geo_info()
        elevations = [cc[2] for cc in self.shapely_geom.coords]
        elev_min = min(elevations)
        elev_max = max(elevations)
        if elev_min == elev_max:
            info['elevation (m)'] = '{:.2f}'.format(elev_min)
        else:
            info['elevation (m)'] = '{:.2f} - {:.2f}'.format(elev_min, elev_max)
        return info


class GeoPolygonModel(GeoModel):
    __abstract__ = True
    shapefile_model: int = POLYGON
    geom: Any = Field(sa_column=Column(Geometry('POLYGON', srid=conf.geo.srid)))
    mapbox_type: str = 'fill'
    base_gis_type: str = 'Polygon'

    @property
    def area(self):
        return transform(reproject_func, self.shapely_geom).area

    @property
    def length(self):
        return transform(reproject_func, self.shapely_geom).length

    @property
    def caption(self):
        """
        Return user friendly name (used in menu, etc)
        :return:
        """
        return '{self.__class__.__name__}: {self.id:d}'.format(self=self)

    @classmethod
    async def get_shapefile_writer(cls):
        writer = ShapeFileWriter(cls.shapefile_model)
        cols = cls.get_columns_for_shapefile_writer(writer)
        query = cls.query.where(cls.geom != None)

        async with db.acquire(reuse=False) as connection:
            async with connection.transaction() as tx:
                ## Add the objects
                async for obj in connection.iterate(query):
                    try:
                        writer.poly(parts=[[point for point in obj.get_points()]])
                    except NoPoint:
                        pass
                    attrs = []
                    for attr_name in cols:
                        attr = getattr(obj, attr_name)
                        if isinstance(attr, date):
                            ## Date in a format approriate for shapefile
                            attrs.append('{:%Y%m%d}'.format(attr))
                        else:
                            attrs.append(attr)
                    writer.record(*attrs)

        return writer

    def get_points(self):
        """
        Get a list of points
        :return:
        """
        if isinstance(self.geom, None.__class__):
            raise NoPoint
        points = wkb.loads(self.geom.data)
        return zip(points.exterior.coords.xy[0], points.exterior.coords.xy[1])

    async def get_geo_info(self):
        info = OrderedDict()
        area = self.area
        bounds = self.shapely_geom.bounds
        info['longitude'] = '{:.6f} - {:.6f}'.format(bounds[0], bounds[2])
        info['latitude'] = '{:.6f} - {:.6f}'.format(bounds[1], bounds[3])
        info['length (m)'] = '{:.2f}'.format(self.length)
        info['area (sq. m)'] = '{:.1f} sq. m'.format(area)
        info['area (ha)'] = '{:.1f} ha'.format(area / 10000)
        info['area (acre)'] = '{:.1f} acres'.format(area / 4046.85643005078874)
        return info


class GeoPolygonModelZ(GeoPolygonModel):
    __abstract__ = True
    shapefile_model: int = POLYGONZ
    geom: Any = Field(sa_column=Column(Geometry('POLYGONZ', dimension=3, srid=conf.geo.srid)))

    async def get_geo_info(self):
        info = await super(GeoPolygonModelZ, self).get_geo_info()
        if hasattr(self.shapely_geom, 'exterior'):
            coords = self.shapely_geom.exterior.coords
        else:
            coords = self.shapely_geom.coords
        elevations = [coord[2] for coord in coords]
        elev_min = min(elevations)
        elev_max = max(elevations)
        if elev_min == elev_max:
            info['elevation (m)'] = '{:.2f}'.format(elev_min)
        else:
            info['elevation (m)'] = '{:.2f} - {:.2f}'.format(elev_min, elev_max)
        return info


class GeoPointSurveyModel(SurveyModel, GeoPointMModel):
    #__abstract__ = True

    ## raw_model is set in category_models_maker.make_category_models
    raw_model: Any = None


class LineWorkSurveyModel(SurveyModel):
    __abstract__ = True

    ## raw_model is set in category_models_maker.make_category_models
    raw_model: Any = None

    def match_raw_points(self):
        reprojected_geom = transform(reproject_func, self.shapely_geom)
        reprojected_geom_geoalchemy = from_shape(reprojected_geom, conf.raw_survey_srid)
        raw_survey_points_project = self.raw_model.query.filter(self.raw_model.project_id==self.project_id)
        query = raw_survey_points_project.filter(
            func.ST_Distance(reprojected_geom_geoalchemy, self.raw_model.geom) < conf.epsilon
        )
        return query.all()


class GeoLineSurveyModel(LineWorkSurveyModel, GeoLineModelZ):
    __abstract__ = True


class GeoPolygonSurveyModel(LineWorkSurveyModel, GeoPolygonModelZ):
    __abstract__ = True


class RawSurveyBaseModel(BaseSurveyModel, GeoPointMModel):
    """
    Abstract base class for category based raw survey point models
    """
    __abstract__ = True
    geom: Any = Field(sa_column=Column(Geometry('POINTZ', dimension=3, srid=conf.geo.raw_survey.srid)))
    status: str = Field(sa_column=Column(String(1)))

    ## store_name is set in category_models_maker.make_category_models
    store_name: str | None = None

    @classmethod
    async def get_geo_df(cls, *args, **kwargs):
        return await super().get_geo_df(crs=conf.raw_survey['spatial_sys_ref'], *args, **kwargs)


class PlottableModel(Model):
    """
    Model that defines a value for each record, therefore that can be used
    in plots (eg. bar graphs).
    Subclasses might want to define:
    * a unit (default: m)
    * a resampling method, to be applied to all values (unless values is a dict, see below)
    * attributes with numerical types, and:
        * a tuple called values that defines what are these columns
          to be used (the first one being the default)
        * OR an ordereed dict of value => resampling method
    """
    __abstract__ = True

    float_format: str = '%.1f'
    values: dict[Any, Any] = {}

    @classmethod
    async def get_as_dataframe(cls, model_id=None, where=None, **kwargs):
        """
        Get a dataframe for the data.
        It's quite generic, so subclasses might want to subclass this.
        """
        if where is None:
            if model_id is None:
                where_ = None
            else:
                where_ = cls.ref_id == model_id
        else:
            if model_id is None:
                where_ = where
            else:
                where_ = and_(cls.ref_id == model_id, where)

        if where_ is not None:
            df = await cls.get_df(where=where_, **kwargs)
        else:
            df = await cls.get_df(**kwargs)

        return df


class TimePlottableModel(PlottableModel):
    __abstract__ = True

    time: datetime

    @classmethod
    async def get_as_dataframe(cls, model_id=None, with_only_columns=None, **kwargs):
        """
        Get the data as a time-indexed dataframe
        """
        if with_only_columns == None:
            with_only_columns = [val['name'] for val in cls.values]
        if 'time' not in with_only_columns:
            with_only_columns.insert(0, 'time')

        df = await super().get_as_dataframe(model_id=model_id,
                                            with_only_columns=with_only_columns, **kwargs)

        ## Set time as index
        df.set_index('time', drop=True, inplace=True)
        df.sort_index(inplace=True)
        return df
