from datetime import datetime
from csv import writer
from collections import defaultdict
from io import BytesIO, StringIO

from sqlmodel import Field, SQLModel, MetaData, JSON, TEXT, Relationship, Column
import pyproj
from shapely.geometry import Point

from ..config import conf
from .models_base import Model
from .metadata import gisaf_admin

class Project(Model):
    metadata = gisaf_admin

    class Admin:
        menu = 'Other'
        flask_admin_model_view = 'ProjectModelView'

    id: int = Field(default=None, primary_key=True)
    name: str
    contact_person: str
    site: str
    date_approved: datetime
    start_date_planned: datetime
    start_date_effective: datetime
    end_date_planned: datetime
    end_date_effective: datetime

    def __str__(self):
        return '{self.name:s}'.format(self=self)

    def __repr__(self):
        return '<models.Project {self.name:s}>'.format(self=self)

    async def auto_import(self, registry):
        """
        Import the points of the given project to the GIS DB
        in their appropriate models.raw_survey_models
        :return: dict of result (stats)
        """
        from .category import Category
        from .models_base import GeoPointSurveyModel

        result = defaultdict(int)

        categories = {cat.table_name: cat for cat in await Category.query.gino.all()}

        ## Define projections
        survey_proj = pyproj.Proj(**conf.raw_survey['spatial_sys_ref'])
        target_proj = pyproj.Proj(f'epsg:{conf.srid:d}')
        def reproject(x, y, z):
            return pyproj.transform(survey_proj, target_proj, x, y, z)

        ## TODO: Gino session
        for survey_model_name, raw_survey_model in registry.raw_survey_models.items():
            category = categories[survey_model_name]
            if not category.auto_import:
                continue
            survey_model = registry.geom_auto.get(survey_model_name)
            if not survey_model:
                continue
            if not issubclass(survey_model, GeoPointSurveyModel):
                continue
            raw_survey_items = await raw_survey_model.query.where(raw_survey_model.project_id == self.id).gino.all()
            for item in raw_survey_items:
                if not item:
                    continue
                new_point = survey_model(
                    id=item.id,
                    date=item.date,
                    accur_id=item.accur_id,
                    equip_id=item.equip_id,
                    srvyr_id=item.srvyr_id,
                    orig_id=item.orig_id,
                    status=item.status,
                    project_id=self.id,
                )
                geom = Point(*reproject(item.easting, item.northing, item.elevation))
                new_point.geom = 'SRID={:d};{:s}'.format(conf.srid, geom.wkb_hex)
                ## TODO: merge with Gino
                #session.merge(new_point)
                result[survey_model_name] += 1
        #session.commit()
        return result


    async def download_raw_survey_data(self):
        from .raw_survey import RawSurveyModel
        ## FIXME: old query style
        breakpoint()
        raw_survey_items = await RawSurveyModel.query.where(RawSurveyModel.project_id == self.id).gino.all()
        csv_file = StringIO()
        csv_writer = writer(csv_file)
        for item in raw_survey_items:
            csv_writer.writerow(item.to_row())
        now = '{:%Y-%m-%d_%H:%M}'.format(datetime.now())

        ## XXX: not tested (aiohttp)
        #return send_file(BytesIO(bytes(csv_file.getvalue(), 'utf-8')),
        #                 attachment_filename='{:s}-{:s}.csv'.format(self.name, now),
        #                 mimetype='text/csv',
        #                 as_attachment=True)
        headers = {
            'Content-Disposition': 'attachment; filename="{}"'.format('{:s}-{:s}.csv'.format(self.name, now))
        }
        return web.Response(
            status=200,
            headers=headers,
            content_type='text/csv',
            body=BytesIO(bytes(csv_file.getvalue(), 'utf-8'))
        )

    async def download_reconciled_raw_survey_data(self, registry):
        csv_file = StringIO()
        csv_writer = writer(csv_file)
        for model_name, model in registry.raw_survey_models.items():
            survey_items = await model.query.where(model.project_id == self.id).gino.all()
            for item in survey_items:
                csv_writer.writerow(item.to_row())
        now = '{:%Y-%m-%d_%H:%M}'.format(datetime.now())

        ## XXX: not tested (aiohttp)
        #return send_file(BytesIO(bytes(csv_file.getvalue(), 'utf-8')),
        #                 attachment_filename='{:s}-{:s}-reconciled.csv'.format(self.name, now),
        #                 mimetype='text/csv',
        #                 as_attachment=True)
        headers = {
            'Content-Disposition': 'attachment; filename="{}"'.format('{:s}-{:s}.csv'.format(self.name, now))
        }
        return web.Response(
            status=200,
            headers=headers,
            content_type='text/csv',
            body=BytesIO(bytes(csv_file.getvalue(), 'utf-8'))
        )

    async def reconcile(self, registry):
        from gisaf.models.reconcile import Reconciliation
        result = {}
        all_reconciliations = await Reconciliation.query.gino.all()
        point_ids_to_reconcile = {p.id: registry.raw_survey_models[p.target]
                                  for p in all_reconciliations
                                  if p.target in registry.raw_survey_models}

        result['bad target'] = set([p.target for p in all_reconciliations
                                    if p.target not in registry.raw_survey_models])
        result['from'] = defaultdict(int)
        result['to'] = defaultdict(int)
        result['unchanged'] = defaultdict(int)

        ## TODO: Gino session
        for model_name, model in registry.raw_survey_models.items():
            points_to_reconcile = await model.query.\
                where(model.project_id==self.id).\
                where(model.id.in_(point_ids_to_reconcile.keys())).gino.all()
            for point in points_to_reconcile:
                new_model = point_ids_to_reconcile[point.id]
                if new_model == model:
                    result['unchanged'][model] += 1
                    continue
                new_point = new_model(
                    id=point.id,
                    accur_id=point.accur_id,
                    srvyr_id=point.accur_id,
                    project_id=point.project_id,
                    status=point.status,
                    orig_id=point.orig_id,
                    equip_id=point.equip_id,
                    geom=point.geom,
                    date=point.date
                )
                ## TODO: Gino add and delete
                #session.add(new_point)
                #session.delete(point)
                result['from'][point.__class__] += 1
                result['to'][new_point.__class__] += 1
        return result



    # def download_raw_survey_data(self, session=None):
    #     from gisaf.models.raw_survey_models import RawSurvey
    #     from gisaf.registry import registry
    #     if not session:
    #         session = db.session
    #     raw_survey_items = session.query(RawSurvey).filter(RawSurvey.project_id == self.id).all()
    #     csv_file = StringIO()
    #     csv_writer = writer(csv_file)
    #
    #     SURVEY_PROJ = pyproj.Proj(**conf.raw_survey['spatial_sys_ref'])
    #     TARGET_PROJ = pyproj.Proj(init='epsg:{:d}'.format(conf.srid))
    #
    #     def reproject(x, y, z):
    #         return pyproj.transform(SURVEY_PROJ, TARGET_PROJ, x, y, z)
    #
    #     for item in raw_survey_items:
    #         csv_writer.writerow(item.to_row())
    #
    #         ## Add import of points, incl. reprojection, to registry.raw_survey_models:
    #         new_coords = reproject(item.easting, item.northing, item.elevation)
    #         geom = Point(*new_coords)
    #         ## TODO: from here
    #         model = registry.raw_survey_models
    #         new_point = model(
    #             id=item.id,
    #             category=item.category,
    #             date=item.date,
    #             accur_id=item.accur_id,
    #             equip_id=item.equip_id,
    #             srvyr_id=item.srvyr_id,
    #             orig_id=item.original_id,
    #         )
    #         new_point.geom = 'SRID={:d};{:s}'.format(conf.srid, geom.wkb_hex)
    #         session.merge(new_point)
    #         result[item.category_info] += 1
    #
    #     now = '{:%Y-%m-%d_%H:%M}'.format(datetime.now())
    #
    #     return send_file(BytesIO(bytes(csv_file.getvalue(), 'utf-8')),
    #                      attachment_filename='{:s}-{:s}.csv'.format(self.name, now),
    #                      mimetype='text/csv',
    #                      as_attachment=True)

